#lang racket
(define (foldr:1 step base-case l)
  (cond
    [(empty? l) base-case]
    [else (step (first l) (foldr:1 step base-case (rest l)))]))

(define (foldr:2 step base-case l)
  (define (foldr-iter accum l)
    (cond
      [(empty? l) (accum base-case)]
      [else
        (foldr-iter
          (lambda (x)
            (accum (step (first l) x)))
          (rest l))]))
  (foldr-iter (lambda (x) x) l))

(define (foldl step base-case l)
  (cond
    [(empty? l) base-case]
    [else (foldl step
            (step (first l) base-case)
            (rest l))]))

(define (map:1 f l)
  (define (on-elem elem new-list)
    (cons (f elem) new-list))
  (foldr:1 on-elem empty l))

(define (map:2 f l)
  (define (on-elem elem new-list)
    (cons (f elem) new-list))
  (foldr:2 on-elem empty l))

(define (map:3 f l)
  (define (on-elem elem new-list)
    (cons (f elem) new-list))
  (foldr:3 on-elem empty l))

(define (filter to-keep? l)
  (define (on-elem elem new-list)
    (cond [(to-keep? elem) (cons elem new-list)]
          [else new-list]))
  (foldr on-elem empty l))

(define (append l r)
  (foldr cons r l))

(define (reverse l)
  (foldl cons empty l))

(define (foldr:3 step base-case l)
  (foldl step base-case (reverse l)))

(map:1 (lambda (x) x) (list 1 2 3))
(map:2 (lambda (x) x) (list 1 2 3))
(map:3 (lambda (x) x) (list 1 2 3))

#|
(filter:2 even? (list 1 2 3 4 5))
(append:2 (list 1 2) (list 3 4))
(reverse:2 (list 1 2 3 4))
|#

; Creates a list with 1000 elements
(define (countup-from1 to)
  (define (count from)
    (cond [(equal? from to) (list to)]
          [else (cons from (count (+ 1 from)))]))
  (count 1))
; A list with 1 million entries
(define (average l)
  (define total (foldl + 0.0 l))
  (define count (length l))
  (/ total count))
(require racket/format)
(define size 1000000)
(define l (sort (build-list size (lambda (x) (random 9999999))) <))
(require "benchmark.rkt")


(printf "Processing a list of size: ~a\n" size)
(define s1 (benchmark (thunk (map:1 (lambda (x) x) l))))

(define (print-sample label s)
  (printf "-------------------------------------------------------\n")
  (printf "Throughoutput (~a): ~a elems/ms \n" label (~r (/ size (sample-arith-mean s)) #:precision 0))
  (printf "Mean (~a): ~ams\n" label (sample->string s)))
(print-sample "unopt" s1)

(define s2 (benchmark (thunk (map:2 (lambda (x) x) l))))
#|
(printf "Throughoutput (tail-rec): ~a elems/ms \n" (~r (/ size (sample-arith-mean s2)) #:precision 0))
(printf "Mean (tail-rec): ~ams\n" (sample->string s2))
|#
(define (print-speedup label s [precision 2])
  (print-sample label s)
  (printf "Speed-up (~a): ~a\n" label (~r (/ (sample-arith-mean s1) (sample-arith-mean s)) #:precision 1)))
(print-speedup "tailrec" s2)


(define s3 (benchmark (thunk (map:3 (lambda (x) x) l))))
(print-speedup "rev+foldl" s3)
